﻿using Barcode.Host.Server.Models;
using Barcode.Host.Shared.Models;
using Microsoft.AspNetCore.Builder;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Configuration;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Hosting;
using Serilog;
using System.Reflection;

namespace Barcode.Host.Server;

public class Program
{

    private static (string, WebApplicationOptions) Get(string[] args)
    {
        Assembly assembly = Assembly.GetExecutingAssembly();
        string? assemblyName = assembly.GetName()?.Name;
        if (string.IsNullOrEmpty(assemblyName))
            throw new Exception();
        string baseAssemblyName = assemblyName.Split('.')[0];
        string webRootPath = Path.Combine(AppContext.BaseDirectory.Split(baseAssemblyName)[0], baseAssemblyName, "wwwroot");
        if (!Directory.Exists(webRootPath))
            webRootPath = string.Empty;
        WebApplicationOptions webApplicationOptions = new()
        {
            Args = args,
            ContentRootPath = AppContext.BaseDirectory,
            WebRootPath = webRootPath
        };
        return new(assemblyName, webApplicationOptions);
    }

    public static int Main(string[] args)
    {
        LoggerConfiguration loggerConfiguration = new();
        (string assemblyName, WebApplicationOptions _) = Get(args);
        WebApplicationBuilder webApplicationBuilder = WebApplication.CreateBuilder(args);
        _ = webApplicationBuilder.Configuration.AddUserSecrets<Program>();
        AppSettings appSettings = Models.Binder.AppSettings.Get(webApplicationBuilder.Configuration);
        if (string.IsNullOrEmpty(appSettings.WorkingDirectoryName))
            throw new Exception("Working directory name must have a value!");
        string workingDirectory = IWorkingDirectory.GetWorkingDirectory(assemblyName, appSettings.WorkingDirectoryName);
        Environment.SetEnvironmentVariable(nameof(workingDirectory), workingDirectory);
        _ = ConfigurationLoggerConfigurationExtensions.Configuration(loggerConfiguration.ReadFrom, webApplicationBuilder.Configuration);
        _ = SerilogHostBuilderExtensions.UseSerilog(webApplicationBuilder.Host);
        Log.Logger = loggerConfiguration.CreateLogger();
        ILogger log = Log.ForContext<Program>();
        try
        {
            _ = webApplicationBuilder.Services.AddMemoryCache();
            _ = webApplicationBuilder.Services.Configure<ApiBehaviorOptions>(options => options.SuppressModelStateInvalidFilter = true);
            _ = webApplicationBuilder.Services.AddControllersWithViews();
            _ = webApplicationBuilder.Services.AddDistributedMemoryCache();

            _ = webApplicationBuilder.Services.AddSingleton(_ => appSettings);

            _ = webApplicationBuilder.Services.AddSwaggerGen();
            _ = webApplicationBuilder.Services.AddSession(sessionOptions =>
                {
                    sessionOptions.IdleTimeout = TimeSpan.FromSeconds(2000);
                    sessionOptions.Cookie.HttpOnly = true;
                    sessionOptions.Cookie.IsEssential = true;
                }
            );
            WebApplication webApplication = webApplicationBuilder.Build();
            _ = webApplication.UseCors(corsPolicyBuilder => corsPolicyBuilder.AllowAnyOrigin().AllowAnyHeader().AllowAnyMethod());
            if (!webApplicationBuilder.Environment.IsDevelopment())
            {
                _ = webApplication.UseExceptionHandler("/Error");
                _ = webApplication.UseHttpsRedirection();
                _ = webApplication.UseHsts();
            }
            else
            {
                if (string.IsNullOrEmpty(appSettings.URLs))
                {
                    Environment.ExitCode = -1;
                    webApplication.Lifetime.StopApplication();
                }
                _ = webApplication.UseSwagger();
                _ = webApplication.UseSwaggerUI(c => c.SwaggerEndpoint("/swagger/v1/swagger.json", "Server V1"));
            }
            _ = webApplication.Lifetime.ApplicationStopped.Register(Log.CloseAndFlush);
            _ = SerilogApplicationBuilderExtensions.UseSerilogRequestLogging(webApplication);
            _ = webApplication.UseFileServer(enableDirectoryBrowsing: true);
            _ = webApplication.UseStaticFiles();
            _ = webApplication.UseSession();
            _ = webApplication.MapControllers();
            log.Information("Starting Web Application");
            webApplication.Run();
            return 0;
        }
        catch (Exception ex)
        {
            log.Fatal(ex, "Host terminated unexpectedly");
            return 1;
        }
        finally
        {
            Log.CloseAndFlush();
        }
    }

}